/* *********************************************************************** *
 * XINX                                                                    *
 * Copyright (C) 2010 by Ulrich Van Den Hekke                              *
 * ulrich.vdh@shadoware.org                                                *
 *                                                                         *
 * This program is free software: you can redistribute it and/or modify    *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * This program is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
 * *********************************************************************** */

#include "calcul_distance.h"
#include <QtSql>
#include <QQueue>
#include <iostream>
#include "compresser.h"

CalculDistance::CalculDistance(QStringList paths, QObject *parent) : QObject(parent), m_paths(paths), m_step(0)
{
	m_uuid = QUuid::createUuid();

	init();
}

CalculDistance::CalculDistance(const QString & uuid, QStringList paths, QObject *parent) : QObject(parent), m_paths(paths), m_uuid(uuid), m_step(0)
{
	init();
}

CalculDistance::~CalculDistance()
{
	QSqlDatabase::database(m_uuid, true).close();
	QSqlDatabase::removeDatabase(m_uuid);
}

void CalculDistance::setFilter(const QStringList & value)
{
	m_filters = value;
}

void CalculDistance::init()
{
	m_singleFile.watcher = 0;
	m_pairFile.watcher = 0;

	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", m_uuid);
	db.setDatabaseName(m_uuid + ".db");
	db.open();
	db.setConnectOptions("QSQLITE_BUSY_TIMEOUT=1000;");
	db.exec("PRAGMA synchronous=OFF");
	db.exec("PRAGMA temp_store=MEMORY");
}

void CalculDistance::singleStepFinish()
{
	delete m_singleFile.watcher;
	m_singleFile.watcher = 0;
	stepFinish();
}

void CalculDistance::pairStepFinish()
{
	delete m_pairFile.watcher;
	m_pairFile.watcher = 0;
	stepFinish();
}

void CalculDistance::stepFinish()
{
	std::cout << "\r                                                                  ";
	std::cout << "\r";
	m_step++;
	process();
}

void CalculDistance::process()
{
	switch(m_step)
	{
	case 0:
		std::cout << "Step 0 : Create database" << std::endl;
		step0_createbase();
		m_step++;
	case 1:
		std::cout << "Step 1 : Create file list" << std::endl;
		step1_createfilelist();
		m_step++;
	case 2:
		std::cout << "Step 2 : Compress single file                    " << std::endl;
		step2_compresssingle();
		break;
	case 3:
		std::cout << "Step 3 : Compress pair file                      " << std::endl;
		step3_compresspair();
		break;
	case 4:
		qApp->quit();
	 }
}

void CalculDistance::step0_createbase()
{
	QSqlDatabase db = QSqlDatabase::database(m_uuid, true);

	if(!db.tables().contains("files"))
	{
		db.exec("CREATE TABLE files(id INTEGER PRIMARY KEY, path TEXT, size INTEGER, compressed_size INTEGER)");
	}
	if(!db.tables().contains("distances"))
	{
		db.exec("CREATE TABLE distances(id1 INTEGER, id2 INTEGER, compressed_size INTEGER, distance INTEGER)");
		db.exec("CREATE UNIQUE INDEX distances_idx1 ON distances (id1 ASC, id2 ASC)");
	}
}

void CalculDistance::step1_createfilelist()
{
	QSqlDatabase db = QSqlDatabase::database(m_uuid, true);

	std::cout << "\r" << tr("Init list of files ...").toStdString();

	QSqlQuery insertFile("INSERT INTO files(path, size) VALUES(?, ?)", db);
	QQueue<QString> queue;
	queue.append(m_paths);

	db.transaction();
	while(!queue.isEmpty())
	{
		QDir directory(queue.dequeue());

		QFileInfoList fileInfoList = directory.entryInfoList(m_filters, QDir::AllDirs | QDir::AllEntries | QDir::NoDotAndDotDot | QDir::NoSymLinks);
		foreach(const QFileInfo & fileInfo, fileInfoList)
		{
			if(fileInfo.isDir())
			{
				queue.enqueue(fileInfo.absoluteFilePath());
			}
			else
			{
				insertFile.bindValue(0, fileInfo.absoluteFilePath());
				insertFile.bindValue(1, fileInfo.size());
				bool result = insertFile.exec();
				Q_ASSERT_X(result, "CalculDistance::step1_createfilelist", qPrintable(insertFile.lastError().text()));
			}
		}
	}
	std::cout << "\r" << tr("Commit files ...").toStdString();
	db.commit();

	std::cout << "\r" << tr("Init list of distances ...").toStdString();
	db.exec("INSERT INTO distances(id1, id2) SELECT files1.id as id1, files2.id as id2 FROM files files1, files files2 WHERE files2.id > files1.id AND NOT EXISTS (SELECT 1 FROM distances WHERE id1=files1.id and id2=files2.id)");

	std::cout << "\r                                                                  ";
	std::cout << "\r";
}

void CalculDistance::singleStepReady(int index)
{
	QSqlDatabase db = QSqlDatabase::database(m_uuid, true);

	QSqlQuery updateQuery("UPDATE files SET compressed_size=? WHERE id=?", db);

	m_singleFile.sum_size            += m_singleFile.watcher->resultAt(index).size;
	m_singleFile.sum_compressed_size += m_singleFile.watcher->resultAt(index).compressed_size;

	updateQuery.bindValue(0, m_singleFile.watcher->resultAt(index).compressed_size);
	updateQuery.bindValue(1, m_singleFile.watcher->resultAt(index).id);
	updateQuery.exec();

	std::cout << "\r" <<
		tr("Process %1/%2 : %3Mo/%4Mo")
			.arg(m_singleFile.watcher->progressValue())
			.arg(m_singleFile.watcher->progressMaximum())
			.arg((double)m_singleFile.sum_size / (1024*1024), 0, 'f', 2)
			.arg((double)m_singleFile.sum_compressed_size / (1024*1024), 0, 'f', 2).toStdString() << "                   ";

}

static ulong getCompressedSize(const QString file1, const QString file2 = QString())
{
	QStringList list;
	list << file1;
	if(!file2.isEmpty())
		list << file2;

	Compresser c(list);
	c.process();
	return c.m_finalSize;
}

static CalculDistance::single_file calculateSingleCompressedSize(const CalculDistance::single_file & file)
{
	CalculDistance::single_file result = { file.id, file.filename, file.size, 0 };
	result.compressed_size = getCompressedSize(file.filename);
	return result;
}

void CalculDistance::step2_compresssingle()
{
	std::cout << "\r" << tr("Init. structure ...").toStdString();

	m_singleFile.watcher = new QFutureWatcher<single_file>(this);
	connect(m_singleFile.watcher, SIGNAL(resultReadyAt(int)), this, SLOT(singleStepReady(int)));
	connect(m_singleFile.watcher, SIGNAL(finished()), this, SLOT(singleStepFinish()));

	QSqlDatabase db = QSqlDatabase::database(m_uuid, true);

	QSqlQuery countQuery("SELECT count(1) FROM files WHERE compressed_size is null", db);
	QSqlQuery selectQuery("SELECT id, path, size FROM files WHERE compressed_size is null", db);

	countQuery.first();
	int count = countQuery.value(0).toInt();
	std::cout << " (" << count << ")";

	QVector<single_file> list;
	list.reserve(count);

	m_singleFile.sum_size = 0;
	m_singleFile.sum_compressed_size = 0;

	while(selectQuery.next())
	{
		struct single_file f;
		f.id       = selectQuery.value(0).toInt();
		f.filename = selectQuery.value(1).toString();
		f.size     = selectQuery.value(2).toULongLong();
		list.append(f);
	}

	std::cout << "\r" << tr("Launch ...").toStdString();
	m_singleFile.watcher->setFuture(QtConcurrent::mapped(list, calculateSingleCompressedSize));
}

void CalculDistance::pairStepReady(int index)
{
	QSqlDatabase db = QSqlDatabase::database(m_uuid, true);

	QSqlQuery updateQuery("UPDATE distances set compressed_size=?, distance=? WHERE id1=? and id2=?", db);

	CalculDistance::pair_file pair = m_pairFile.watcher->resultAt(index);

	m_pairFile.sum_size            += pair.files->value(pair.id1).size + pair.files->value(pair.id2).size;
	m_pairFile.sum_compressed_size += pair.compressed_size;

	updateQuery.bindValue(0, pair.compressed_size);
	updateQuery.bindValue(1, pair.distance);
	updateQuery.bindValue(2, pair.id1);
	updateQuery.bindValue(3, pair.id2);
	updateQuery.exec();

	uint progressValue = m_pairFile.watcher->progressValue();
	uint progressMaximum = m_pairFile.watcher->progressMaximum();

	uint elapsedTime = QDateTime::currentDateTime().toTime_t() - m_start;
	uint workingTime = (progressMaximum * elapsedTime) / progressValue;
	uint elapsedTimeH, elapsedTimeM, elapsedTimeS;
	uint workingTimeH, workingTimeM, workingTimeS;

	elapsedTimeS = elapsedTime % 60; elapsedTime = elapsedTime / 60;
	elapsedTimeM = elapsedTime % 60; elapsedTime = elapsedTime / 60;
	elapsedTimeH = elapsedTime;

	workingTimeS = workingTime % 60; workingTime = workingTime / 60;
	workingTimeM = workingTime % 60; workingTime = workingTime / 60;
	workingTimeH = workingTime;

	QString sElapsedTime = QString("%1:%2:%3").arg(elapsedTimeH, 2, 10, QChar('0')).arg(elapsedTimeM, 2, 10, QChar('0')).arg(elapsedTimeS, 2, 10, QChar('0'));
	QString sWorkingTime = QString("%1:%2:%3").arg(workingTimeH, 2, 10, QChar('0')).arg(workingTimeM, 2, 10, QChar('0')).arg(workingTimeS, 2, 10, QChar('0'));

	std::cout << "\r" <<
		tr("Process %1/%2 : %3Mo/%4Mo (%5 / %6)")
			.arg(progressValue)
			.arg(progressMaximum)
			.arg((double)m_pairFile.sum_size / (1024*1024), 0, 'f', 2)
			.arg((double)m_pairFile.sum_compressed_size / (1024*1024), 0, 'f', 2)
			.arg(sElapsedTime)
			.arg(sWorkingTime).toStdString() << "                                                          ";

}

static CalculDistance::pair_file calculatePairCompressedSize(const CalculDistance::pair_file & pair)
{
	CalculDistance::pair_file result = { pair.files, pair.id1, pair.id2, 0, 0.0 };
	const CalculDistance::single_file file1 = pair.files->value(pair.id1);
	const CalculDistance::single_file file2 = pair.files->value(pair.id2);

	double compressed1 = file1.compressed_size;
	double compressed2 = file2.compressed_size;
	double max_size    = (compressed1 > compressed2 ? compressed1 : compressed2);

	result.compressed_size = getCompressedSize(file1.filename, file2.filename);
	result.distance        = 1 - (compressed1 + compressed2 - result.compressed_size) / max_size;

	return result;
}

void CalculDistance::step3_compresspair()
{
	std::cout << "\r" << tr("Init. files structure ...").toStdString();

	m_pairFile.watcher = new QFutureWatcher<pair_file>(this);
	connect(m_pairFile.watcher, SIGNAL(resultReadyAt(int)), this, SLOT(pairStepReady(int)));
	connect(m_pairFile.watcher, SIGNAL(finished()), this, SLOT(pairStepFinish()));

	QSqlDatabase db = QSqlDatabase::database(m_uuid, true);

	m_pairFile.files.clear();
	QSqlQuery selectFilesQuery("select id, path, size, compressed_size from files", db);
	while (selectFilesQuery.next())
	{
		single_file sf;

		sf.id              = selectFilesQuery.value(0).toInt();
		sf.filename        = selectFilesQuery.value(1).toString();
		sf.size            = selectFilesQuery.value(2).toInt();
		sf.compressed_size = selectFilesQuery.value(3).toInt();

		m_pairFile.files.insert(sf.id, sf);
	}

	std::cout << "\r" << tr("Init. distances structure ...").toStdString();

	QSqlQuery countDistanceQuery("select count(1) from distances where distance is null", db);
	QSqlQuery selectDistanceQuery("select id1, id2 from distances where distances.distance is null", db);

	countDistanceQuery.first();
	int count = countDistanceQuery.value(0).toInt();
	std::cout << " (" << count << ")";

	QVector<pair_file> list;
	list.reserve(count);

	m_pairFile.sum_size = 0;
	m_pairFile.sum_compressed_size = 0;

	while(selectDistanceQuery.next())
	{
		pair_file p;

		p.files = &(m_pairFile.files);
		p.id1 = selectDistanceQuery.value(0).toInt();
		p.id2 = selectDistanceQuery.value(1).toInt();
		p.distance = 0;
		p.compressed_size = 0;

		list.append(p);
	}


	std::cout << "\r" << tr("Launch ...").toStdString();
	m_start = QDateTime::currentDateTime().toTime_t();
	m_pairFile.watcher->setFuture(QtConcurrent::mapped(list, calculatePairCompressedSize));
}

