/* *********************************************************************** *
 * XINX                                                                    *
 * Copyright (C) 2010 by Ulrich Van Den Hekke                              *
 * ulrich.vdh@shadoware.org                                                *
 *                                                                         *
 * This program is free software: you can redistribute it and/or modify    *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * This program is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
 * *********************************************************************** */

#include "compresser.h"
#include <QCoreApplication>
#include <QFile>
#include <QDebug>

extern "C" {
#include "lzma465/C/LzmaEnc.h"
}


static void * AllocForLzma(void *, size_t size) { return malloc(size); }
static void FreeForLzma(void *, void *address) { free(address); }
static ISzAlloc SzAllocForLzma = { &AllocForLzma, &FreeForLzma };

typedef struct
{
	ISeqInStream SeqInStream;
	QIODevice * device;
	Compresser * m_compresser;
} VectorInStream;

SRes VectorInStream_Read(void *p, void *buf, size_t *size)
{
	VectorInStream *ctx = (VectorInStream*)p;

	if (ctx->device == 0)
	{
		if (ctx->m_compresser->m_index == ctx->m_compresser->m_paths.size())
		{
			*size = 0;
			return SZ_OK;
		}

		const QString path = ctx->m_compresser->m_paths.at(ctx->m_compresser->m_index);
		QFile * device = new QFile(path);
		if (!device->open(QIODevice::ReadOnly))
		{
			*size = 0;
			qWarning() << qApp->translate("Compresser", "Can't open file name `%1`.").arg(path);
			return SZ_OK;
		}
		ctx->device = device;
		ctx->m_compresser->m_index++;
	}

	size_t newSize = ctx->device->read((char*)buf, *size);

	if (newSize == 0)
	{
		delete ctx->device;
		ctx->device = 0;
		return VectorInStream_Read(p, buf, size);
	}
	else
	{
		*size = newSize;
	}

	return SZ_OK;
}

typedef struct
{
	ISeqOutStream SeqOutStream;
	quint64 * m_size;
} VectorOutStream;

size_t VectorOutStream_Write(void *p, const void *, size_t size)
{
  VectorOutStream *ctx = (VectorOutStream*)p;
  if (size)
  {
	*(ctx->m_size) += size;
  }
  return size;
}

Compresser::Compresser(QStringList paths) : m_paths(paths)
{

}

void Compresser::process()
{
	m_index = 0;
	m_finalSize = 0;

	CLzmaEncHandle enc = LzmaEnc_Create(&SzAllocForLzma);
	Q_ASSERT_X(enc, "Compresser::process", "Can't LzmaEnc_Create");

	CLzmaEncProps props;
	LzmaEncProps_Init(&props);
	//props.level        = 9;

	SRes res = LzmaEnc_SetProps(enc, &props);
	Q_ASSERT_X(res == SZ_OK, "Compresser::process", "Can't LzmaEnc_SetProps");

	size_t propsSize = LZMA_PROPS_SIZE;
	unsigned char propsOut[LZMA_PROPS_SIZE];

	res = LzmaEnc_WriteProperties(enc, propsOut, &propsSize);
	Q_ASSERT_X(res == SZ_OK && propsSize == LZMA_PROPS_SIZE, "Compresser::process", "Can't LzmaEnc_WriteProperties");

	VectorInStream inStream = { { &VectorInStream_Read }, 0, this };
	VectorOutStream outStream = { { &VectorOutStream_Write }, &m_finalSize };

	res = LzmaEnc_Encode(enc, (ISeqOutStream*)&outStream, (ISeqInStream*)&inStream, 0, &SzAllocForLzma, &SzAllocForLzma);
	Q_ASSERT_X(res == SZ_OK, "Compresser::process", "Can't LzmaEnc_Encode");

	LzmaEnc_Destroy(enc, &SzAllocForLzma, &SzAllocForLzma);
}

quint64 Compresser::getFinalSize()
{
	return m_finalSize;
}
